"""Test vocab"""

import os

from collections import defaultdict

import cf_pandas as cfp


def test_init():
    vocab = cfp.Vocab()
    assert isinstance(vocab.vocab, defaultdict)


def test_make_entry():
    vocab = cfp.Vocab()
    vocab.make_entry("temp", ["a", "b"], attr="name")
    assert vocab.vocab == {"temp": {"name": "a|b"}}


def test_add_vocabs():
    vocab = cfp.Vocab()
    vocab.make_entry("temp", ["a", "b"], attr="standard_name")
    vocab.make_entry("salt", ["a", "b"], attr="name")
    compare = {"temp": {"standard_name": "a|b|a|b"}, "salt": {"name": "a|b|a|b"}}
    assert (vocab + vocab).vocab == compare

    vocab2 = cfp.Vocab()
    vocab2.make_entry("temp", ["a", "b"], attr="name")
    compare = {
        "temp": {"name": "a|b", "standard_name": "a|b|a|b"},
        "salt": {"name": "a|b|a|b"},
    }
    assert (vocab + vocab2).vocab == compare


def test_make_more_entries():
    vocab = cfp.Vocab()
    vocab.make_entry("temp", ["a", "b"], attr="name")
    vocab.make_entry("temp", ["a", "b"], attr="standard_name")
    vocab.make_entry("salt", ["a", "b"], attr="name")
    vocab.make_entry("salt", ["a", "b"], attr="name")
    compare = {
        "temp": {"name": "a|b", "standard_name": "a|b"},
        "salt": {"name": "a|b|a|b"},
    }
    assert vocab.vocab == compare


def test_make_entry_with_Reg():
    reg1 = cfp.Reg(
        exclude=["abc", "123"],
        exclude_end=["def", "45"],
        exclude_start=["ghi", "67"],
        include=["jkl", "mno"],
        include_end="z",
        include_start="the",
    )
    vocab = cfp.Vocab()
    vocab.make_entry("words", reg1.pattern(), attr="name")
    assert vocab.vocab == {"words": {"name": reg1.pattern()}}
    reg2 = cfp.Reg(
        exclude=["abc", "45"],
        exclude_end=["def", "67"],
        exclude_start=["ghi", "123"],
        include=["jkl", "pr"],
        include_end="z",
        include_start="the",
    )
    vocab.make_entry("words", reg2.pattern(), attr="name")
    assert vocab.vocab == {"words": {"name": f"{reg1.pattern()}|{reg2.pattern()}"}}


def test_save_and_open(tmpdir):
    # save
    vocab = cfp.Vocab()
    vocab.make_entry("temp", ["a", "b"], attr="name")
    fname = f"{tmpdir}/testsave.json"
    vocab.save(fname)
    assert os.path.exists(fname) == 1

    # open
    vocab2 = cfp.Vocab(fname)
    assert vocab.vocab == vocab2.vocab



"""Test cf-pandas."""

import pandas as pd
import pytest

import cf_pandas as cfp


criteria = {
    "wind_s": {
        "standard_name": "wind_speed$",
    },
}


def test_options():
    # test for inputting a nonexistent option
    with pytest.raises(ValueError):
        cfp.set_options(DISPLAY_WIDTH=80)


def test_match_criteria_key_accessor():

    df = pd.DataFrame(
        columns=[
            "temp",
            "wind_speed",
            "wind_speed (m/s)",
            "WIND_SPEED",
            "wind_speed_status",
        ]
    )

    # test accessor with set_options criteria
    with cfp.set_options(custom_criteria=criteria):
        assert sorted(df.cf["wind_s"].columns) == ["wind_speed", "wind_speed (m/s)"]



"""Test cf-pandas utils."""

import cf_pandas as cfp


criteria = {
    "wind_s": {
        "standard_name": "wind_speed$",
    },
}


def test_match_criteria_key():

    vals = ["wind_speed", "WIND_SPEED", "wind_speed_status"]

    # test function with set_options criteria
    with cfp.set_options(custom_criteria=criteria):
        assert cfp.match_criteria_key(vals, ["wind_s"]) == ["wind_speed"]

    # test function with input criteria
    assert cfp.match_criteria_key(vals, ["wind_s"], criteria) == ["wind_speed"]


def test_match_criteria_key_split():

    vals = ["wind_speed (m/s)", "WIND_SPEED", "wind_speed_status"]

    # test function with set_options criteria
    with cfp.set_options(custom_criteria=criteria):
        assert cfp.match_criteria_key(vals, ["wind_s"], split=True) == [
            "wind_speed (m/s)"
        ]

    # test function with input criteria
    assert cfp.match_criteria_key(vals, ["wind_s"], criteria, split=True) == [
        "wind_speed (m/s)"
    ]


def test_standard_names():

    names = cfp.standard_names()
    assert "sea_water_temperature" in names



"""Class for writing regular expressions."""

from typing import List, Optional, Sequence, Type, Union

from .utils import astype


class Reg(object):
    """Class to write a regular expression.

    Notes
    -----
    * Input strings are never allowed to be empty.
    * Need escape characters on any special characters, and then convert to raw, e.g., r"\[celsius\]" for "[celsius]".
    * The `exclude` options are logical "or".
    * The `include` option is logical "and", `include_or` is logical "or", and the other `include_` options allow for only one selection. If you want to use more than one `include_start` for example, you should make an additional regular expression.
    """

    def __init__(
        self,
        exclude: Optional[Union[List[str], str]] = None,
        exclude_start: Optional[Union[List[str], str]] = None,
        exclude_end: Optional[Union[List[str], str]] = None,
        include: Optional[Union[List[str], str]] = None,
        include_or: Optional[Union[List[str], str]] = None,
        include_exact: Optional[str] = None,
        include_start: Optional[str] = None,
        include_end: Optional[str] = None,
        ignore_case: bool = True,
    ):

        self._exclude = (
            [] if exclude is None or exclude == "" else astype(exclude, list)
        )
        self._exclude_start = (
            []
            if exclude_start is None or exclude_start == ""
            else astype(exclude_start, list)
        )
        self._exclude_end = (
            []
            if exclude_end is None or exclude_start == ""
            else astype(exclude_end, list)
        )
        self._include = (
            [] if include is None or include == "" else astype(include, list)
        )
        self._include_or = (
            [] if include_or is None or include_or == "" else astype(include_or, list)
        )
        self._include_exact = (
            "" if include_exact is None or include_exact == "" else include_exact
        )
        self._include_start = (
            "" if include_start is None or include_start == "" else include_start
        )
        self._include_end = (
            "" if include_end is None or include_end == "" else include_end
        )

        self.ignore_case = ignore_case

        self.check()

    def check(self):
        """Check to make sure selected options are compatible."""

        others = [
            self._exclude,
            self._exclude_start,
            self._exclude_end,
            self._include,
            self._include_or,
            self._include_end,
            self._include_start,
        ]
        if (len(self._include_exact) > 0) and any([len(attr) > 0 for attr in others]):
            raise ValueError(
                "If `include_exact` is used, do not input any other options."
            )

        if not isinstance(self._include_exact, str):
            raise TypeError("`include_exact` should be a str.")

        if not isinstance(self._include_end, str):
            raise TypeError("`include_end` should be a str.")

        if not isinstance(self._include_start, str):
            raise TypeError("`include_start` should be a str.")

    def exclude(self, string: Union[str, list]):
        """Exclude string from anywhere in matches.

        Parameters
        ----------
        string: str, list
            Matches with regular expression `pattern` will not contain string(s).

        Notes
        -----
        As a list of strings, this acts as a logical "or" for the exclusions.
        """

        if string != "":
            self._exclude += astype(string, list)
        self.check()

    def exclude_start(self, string: Union[str, list]):
        """Exclude string from start of matches.

        Parameters
        ----------
        string: str, list
            Matches with regular expression `pattern` will not start with string(s).

        Notes
        -----
        As a list of strings, this acts as a logical "or" for the exclusions.
        """

        if string != "":
            self._exclude_start += astype(string, list)
        self.check()

    def exclude_end(self, string: Union[str, list]):
        """Exclude string from end of matches.

        Parameters
        ----------
        string: str, list
            Matches with regular expression `pattern` will not end with string(s).

        Notes
        -----
        As a list of strings, this acts as a logical "or" for the exclusions.
        """

        if string != "":
            self._exclude_end += astype(string, list)
        self.check()

    def include_exact(self, string: str):
        """String must match exactly.

        Parameters
        ----------
        string: str
            A match with regular expression `pattern` will be exactly string.
        """

        if len(self._include_exact) > 0:
            raise ValueError("`include_exact` already contains a string.")
        if string != "":
            self._include_exact = string
        self.check()

    def include(self, string: Union[str, list]):
        """String must be present anywhere in matches, logical "and".

        Parameters
        ----------
        string: str, list
            Matches with regular expression `pattern` will contain all string(s).

        Notes
        -----
        A list of strings will be treated as a logical "and".
        """

        if string != "":
            self._include += astype(string, list)
        self.check()

    def include_or(self, string: Union[str, list]):
        """String must be present anywhere in matches, logical "or".

        Parameters
        ----------
        string: str, list
            Matches with regular expression `pattern` will contain at lease one of string(s).

        Notes
        -----
        A list of strings will be treated as a logical "or".
        """

        if string != "":
            self._include += astype(string, list)
        self.check()

    def include_end(self, string: str):
        """String must be present at the end of matches.

        Parameters
        ----------
        string: str
            Matches with regular expression `pattern` will end with string.
        """

        if len(self._include_end) > 0:
            raise ValueError("`include_end` already contains a string.")
        if string != "":
            self._include_end = string
        self.check()

    def include_start(self, string: str):
        """String must be present at the start of matches.

        Parameters
        ----------
        string: str
            Matches with regular expression `pattern` will start with string.
        """

        if len(self._include_start) > 0:
            raise ValueError("`include_start` already contains a string.")
        if string != "":
            self._include_start = string
        self.check()

    def pattern(self) -> str:
        """Generate regular expression pattern from user rules.

        Returns
        -------
        str
            Regular expression accounting for all input selections.
        """

        self._pattern = ""

        # the order of these statements is critical to get expressions correct
        if self.ignore_case:
            self._pattern += "(?i)"

        if len(self._exclude) > 0:  # this should be first
            self._pattern += f"^(?!.*({'|'.join(self._exclude)}))"

        if len(self._exclude_start) > 0:
            self._pattern += f"^(?!({'|'.join(self._exclude_start)}))"

        if len(self._exclude_end) > 0:
            self._pattern += f"(?!.*({'|'.join(self._exclude_end)})$)"

        if len(self._include_start) > 0:
            self._pattern += f"^{self._include_start}.*"

        # this is logical "or"
        if len(self._include_or) > 0:
            self._pattern += f".*({'|'.join(self._include_or)}).*"

        # this is logical "and"
        if len(self._include) > 0:
            self._pattern += "".join([f"(?=.*{string})" for string in self._include])

        if len(self._include_end) > 0:
            self._pattern += f".*{self._include_end}$"

        # Exact matches are their own expression, not combined with any others.
        if len(self._include_exact) > 0:
            self._pattern = f"{self._include_exact}$"

        return self._pattern


def joinpat(regs: Sequence[Reg]) -> str:
    """Join patterns from Reg objects.

    Parameters
    ----------
    regs: Sequence
        Reg objects from which `.pattern()` will be used.

    Returns
    -------
    str
        Regular expression patterns from regs joined together with "|"
    """

    return "|".join([reg.pattern() for reg in regs])



"""
From cf-xarray.
"""

import pandas as pd

import cf_pandas as cfp


try:
    # delete the accessor to avoid warning
    del pd.DataFrame.cf
except AttributeError:
    pass


@pd.api.extensions.register_dataframe_accessor("cf")
class CFAccessor:
    """Dataframe accessor analogous to cf-xarray accessor."""

    def __init__(self, pandas_obj):
        # self._validate(pandas_obj)
        self._obj = pandas_obj

    # @staticmethod
    # def _validate(obj):
    #     # verify there is a column latitude and a column longitude
    #     if "latitude" not in obj.columns or "longitude" not in obj.columns:
    #         raise AttributeError("Must have 'latitude' and 'longitude'.")

    def __getitem__(self, key: str):
        """Redefinition of dict-like behavior.
        This enables user to use syntax `reader[dataset_id]` to read in and
        save dataset into the object.
        Parameters
        ----------
        key: str
            dataset_id for a dataset that is available in the search/reader
            object.
        Returns
        -------
        xarray Dataset of the data associated with key
        """

        col_name = cfp.match_criteria_key(self._obj.columns.values, key, split=True)
        return self._obj[col_name]



"""
Utilities for cf-pandas.
"""

from collections import ChainMap
from typing import Any, Iterable, Optional, Union

import pandas as pd
import regex

from .options import OPTIONS


def always_iterable(obj: Any, allowed=(tuple, list, set, dict)) -> Iterable:
    """This is from cf-xarray."""
    return [obj] if not isinstance(obj, allowed) else obj


def astype(value, type_):
    """Return `value` as type `type_`.
    Particularly made to work correctly for returning string, `PosixPath`, or `Timestamp` as list.
    """

    if not isinstance(value, type_):
        import pathlib

        if type_ == list and isinstance(value, (str, pathlib.PurePath, pd.Timestamp)):
            return [value]
        return type_(value)
    return value


def set_up_criteria(criteria: Union[dict, Iterable] = None) -> ChainMap:
    """Get custom criteria from options.

    Parameters
    ----------
    criteria : dict, optional
        Criteria to use to map from variable to attributes describing the variable. If user has defined
        custom_criteria, this will be used by default.

    Returns
    -------
    ChainMap
        Criteria
    """

    if criteria is None:
        if not OPTIONS["custom_criteria"]:
            raise ValueError(
                "criteria needs to be defined either using set_options or directly input."
            )
        criteria_it = OPTIONS["custom_criteria"]
    else:
        criteria_it = always_iterable(criteria, allowed=(tuple, list, set))

    return ChainMap(*criteria_it)


def match_criteria_key(
    available_values: list,
    keys_to_match: Union[str, list],
    criteria: Optional[dict] = None,
    split: bool = False,
) -> list:
    """Use criteria to choose match to key from available available_values.

    Parameters
    ----------
    available_values: list
        String or list of strings to compare against list of category values. They should be keys in `criteria`.
    keys_to_match : str, list
        Key(s) from criteria to match with available_values.
    criteria : dict, optional
        Criteria to use to map from variable to attributes describing the variable. If user has defined custom_criteria, this will be used by default.
    split : bool, optional
        If split is True, split the available_values by white space before performing matches. This is helpful e.g. when columns headers have the form "standard_name (units)" and you want to match standard_name.

    Returns
    -------
    list
        Values from available_values that match keys_to_match, according to criteria.

    Notes
    -----
    This uses logic from `cf-xarray`.
    """

    custom_criteria = set_up_criteria(criteria)

    keys_to_match = astype(keys_to_match, list)
    results = []
    for key in keys_to_match:

        if custom_criteria is not None and key in custom_criteria:
            # criterion is the attribute type — in this function we don't use it,
            # instead we use all the patterns available in criteria to match with available_values
            for criterion, patterns in custom_criteria[key].items():
                if split:
                    results.extend(
                        list(
                            set(
                                [
                                    value
                                    for value in available_values
                                    for value_part in value.split()
                                    if regex.match(patterns, value_part)
                                ]
                            )
                        )
                    )

                else:
                    results.extend(
                        list(
                            set(
                                [
                                    value
                                    for value in available_values
                                    if regex.match(patterns, value)
                                ]
                            )
                        )
                    )

        # catch scenario that user input valid reader variable names
        else:
            if key in available_values:
                results.append(key)
    return list(set(results))


def standard_names():
    """Returns list of CF standard_names.

    Returns
    -------
    list
        All CF standard_names
    """

    import requests
    from bs4 import BeautifulSoup

    url = "https://cfconventions.org/Data/cf-standard-names/79/src/cf-standard-name-table.xml"
    req = requests.get(url)
    soup = BeautifulSoup(req.content, features="xml")

    standard_names = [entry.get("id") for entry in soup.find_all("entry")]

    return standard_names



"""Test widget."""

import pytest

import cf_pandas as cfp


def test_init():
    cfp.Selector(options=["var1", "var2", "var3"])
    cfp.dropdown("temp", ["var1"])


def test_dropdown():
    # exclude var removes "var1"
    w = cfp.dropdown("temp", ["var1"], exclude="var")
    assert w.options == ()

    # include var includes only "var1"
    w = cfp.dropdown("temp", ["var1", "act2"], include="var")
    assert w.options[0] == "var1"


def test_selector():
    # cfp.Selector(options=["var1", "var2", "var3"])

    w = cfp.Selector(
        options=["var1", "var2", "act1", "act2"],
        nickname_in="temp",
        exclude_in="var",
        include_in="1",
    )
    w.button_pressed()  # make entry with default options
    assert w.vocab.vocab == {"temp": {"standard_name": "act1$"}}

    assert w.nickname == w.dropdown.widget.kwargs["nickname"] == "temp"


def test_selector_no_nickname():
    w = cfp.Selector(options=["var1", "var2", "act1", "act2"])
    with pytest.raises(KeyError):  # no nickname
        w.button_pressed()


def test_selector_input_vocab():
    vocab = cfp.Vocab()
    vocab.make_entry("key", ["var"])
    w = cfp.Selector(
        options=["var1", "var2", "act1", "act2"], vocab=vocab, nickname_in="key"
    )

    # initially contains input vocab
    assert w.vocab.vocab == {"key": {"standard_name": "var"}}

    # then make entry and change vocab
    w.button_pressed()  # make entry with default options — so only first valid value is kept
    assert w.vocab.vocab == {"key": {"standard_name": "var|var1$"}}
